from qtpy.QtCore import Qt
from qtpy.QtWidgets import QDoubleSpinBox, QSpinBox, QSlider

from pyqtschema.widgets.base import SchemaWidgetMixin, state_property


def range_from_schema(schema: dict):
    # https://json-schema.org/understanding-json-schema/reference/numeric.html#range
    # there should be an exclusive-minimum OR a minimum
    minimum = schema.get('exclusiveMinimum', schema.get('minimum', None))
    maximum = schema.get('exclusiveMaximum', schema.get('maximum', None))
    return minimum, maximum


class SpinSchemaWidget(SchemaWidgetMixin, QSpinBox):

    @state_property
    def state(self) -> int:
        return self.value()

    @state.setter
    def state(self, state: int):
        self.setValue(state)

    def configure(self):
        self.valueChanged.connect(self.on_changed.emit)

        minimum, maximum = range_from_schema(self.schema)
        if maximum is None:
            maximum = 1_000_000_000
        if maximum:
            self.setMaximum(maximum)
        if minimum:
            self.setMinimum(minimum)

        if "multipleOf" in self.schema:
            self.setSingleStep(self.schema["multipleOf"])


class SpinDoubleSchemaWidget(SchemaWidgetMixin, QDoubleSpinBox):

    @state_property
    def state(self) -> float:
        return self.value()

    @state.setter
    def state(self, state: float):
        self.setValue(state)

    def configure(self):
        self.valueChanged.connect(self.on_changed.emit)

        minimum, maximum = range_from_schema(self.schema)
        if maximum is None:
            maximum = 1_000_000_000
        if maximum:
            self.setMaximum(maximum)
        if minimum:
            self.setMinimum(minimum)

        if "multipleOf" in self.schema:
            self.setSingleStep(self.schema["multipleOf"])


class IntegerRangeSchemaWidget(SchemaWidgetMixin, QSlider):

    def __init__(self, schema: dict, ui_schema: dict, widget_builder: 'WidgetBuilder', *args, **kwargs):
        kwargs['orientation'] = Qt.Horizontal
        super().__init__(schema, ui_schema, widget_builder, *args, **kwargs)

    @state_property
    def state(self) -> int:
        return self.value()

    @state.setter
    def state(self, state: int):
        self.setValue(state)

    def configure(self):
        self.valueChanged.connect(self.on_changed.emit)
        schema = self.schema

        minimum, maximum = range_from_schema(schema)

        multiple_of = schema.get('multipleOf', None)
        if multiple_of:
            self.setTickInterval(multiple_of)
            self.setSingleStep(multiple_of)
            self.setTickPosition(self.TicksBothSides)

        if minimum is None or maximum is None:
            raise ValueError('Cannot create a slider without boundary conditions')
        self.setRange(minimum, maximum)



import sys
from json import load
from pathlib import Path
from typing import Dict, Callable, Optional

from qtpy.QtGui import QIcon
from qtpy.QtWidgets import QApplication, QMainWindow, QMenu, QAction, QScrollArea, QStyle, QMessageBox, QFileDialog

from .__version__ import version
from .builder import WidgetBuilder
from .schema import schema_from_json
from .utils import json_dump


def _exit(txt: str):
    if txt:
        print(txt)
    exit(1)


class MainWindow(QMainWindow):
    def __init__(self, schema: Dict, *args, **kwargs):
        super().__init__(*args, **kwargs)
        style = self.style()

        self.menu_file = QMenu('File', self.menuBar())
        self.menu_help = QMenu('Help', self.menuBar())

        load_action = self._create_action('Load data', triggered=self.load_file,
                                          icon=style.standardIcon(QStyle.SP_FileIcon))
        save_action = self._create_action('Save data', triggered=self.save_file,
                                          icon=style.standardIcon(QStyle.SP_DialogSaveButton))
        quit_action = self._create_action('Quit', triggered=self.close,
                                          icon=style.standardIcon(QStyle.SP_BrowserStop))
        for _act in (load_action, save_action, quit_action):
            self.menu_file.addAction(_act)

        about_action = self._create_action('About', triggered=self.about,
                                           icon=style.standardIcon(QStyle.SP_DialogHelpButton))
        self.menu_help.addAction(about_action)

        self.menuBar().addMenu(self.menu_file)
        self.menuBar().addMenu(self.menu_help)

        ui_schema = {}

        builder = WidgetBuilder(schema)
        self.schema_form = builder.create_form(ui_schema)
        # form.widget.on_changed.connect(lambda d: print(dumps(d, indent=4)))

        widget = QScrollArea()
        widget.setWidget(self.schema_form)
        widget.setWidgetResizable(True)
        self.setCentralWidget(widget)

    def _create_action(self, text: str, icon: Optional[QIcon] = None, tip: Optional[str] = None,
                       toggled: Optional[Callable] = None, triggered: Optional[Callable] = None, ) -> QAction:
        """Create a QAction"""
        action = QAction(text, self)
        if triggered is not None:
            action.triggered.connect(triggered)
        if toggled is not None:
            action.toggled.connect(toggled)
            action.setCheckable(True)
        if icon is not None:
            action.setIcon(icon)
        if tip is not None:
            action.setToolTip(tip)

        return action

    def load_file(self):
        _json_filter = 'json (*.json)'
        f_name = QFileDialog.getOpenFileName(self, 'Load data', '', f'{_json_filter};;All (*)')
        if f_name[0] == '':
            return

        with open(f_name[0], 'r') as stream:
            data = load(stream)
        self.schema_form.state = data

    def save_file(self):
        indent = 2

        _json_filter = 'json (*.json)'
        f_name = QFileDialog.getSaveFileName(self, 'Save data', '', f'{_json_filter};;All (*)')
        if f_name[0] == '':
            return

        filename = f_name[0]
        if f_name[1] == _json_filter and not str(f_name[0]).endswith('.json'):
            filename = f_name[0] + '.json'

        state = self.schema_form.state
        with open(filename, 'w') as OUT:
            json_dump(state, OUT, indent=indent)

    def about(self):
        _text = f'PyQtSchema: {version}'
        QMessageBox.about(self, "pyqtschema", _text)


def run(path_schema: Path, *qt_args):
    schema = schema_from_json(path)

    app = QApplication(*qt_args)

    main = MainWindow(schema)
    main.setWindowTitle(str(path_schema))
    main.show()

    app.exec()


if __name__ == '__main__':
    args = sys.argv[1:]

    if len(args) == 0:
        _exit('Specify a path to a json-schema')

    path = Path(args[0])

    _args = [] if len(args) == 1 else args[1:]

    run(path, _args)



from abc import abstractmethod
from copy import deepcopy
from typing import Dict, Optional, Union

from . import widgets
from .schema import Schema


def get_schema_type(schema: dict) -> str:
    return schema['type']


def has_schema_type(schema: Dict) -> bool:
    return 'type' in schema


def is_enum(schema: dict):
    return 'enum' in schema


def is_any_of(schema: dict):
    return 'anyOf' in schema


def is_all_of(schema: dict):
    return 'allOf' in schema


class IBuilder:
    """ Defines a builder-interface """

    @abstractmethod
    def create_widget(self, schema: dict, ui_schema: dict, state=None) -> 'SchemaWidgetMixin':
        raise NotImplementedError()


class WidgetBuilder(IBuilder):
    default_widget_map = {
        "boolean": {"checkbox": widgets.CheckboxSchemaWidget, "enum": widgets.EnumSchemaWidget},
        "object": {"object": widgets.ObjectSchemaWidget, "enum": widgets.EnumSchemaWidget},
        "number": {"spin": widgets.SpinDoubleSchemaWidget, "text": widgets.TextSchemaWidget,
                   "enum": widgets.EnumSchemaWidget},
        "string": {"textarea": widgets.TextAreaSchemaWidget, "text": widgets.TextSchemaWidget,
                   "password": widgets.PasswordWidget,
                   "path": widgets.FilepathSchemaWidget, "filepath": widgets.FilepathSchemaWidget,
                   "dirpath": widgets.DirectoryPathSchemaWidget,
                   "directory-path": widgets.DirectoryPathSchemaWidget,
                   "colour": widgets.ColorSchemaWidget, "enum": widgets.EnumSchemaWidget},
        "integer": {"spin": widgets.SpinSchemaWidget, "text": widgets.TextSchemaWidget,
                    "range": widgets.IntegerRangeSchemaWidget,
                    "enum": widgets.EnumSchemaWidget},
        "array": {"array": widgets.ArraySchemaWidget, "enum": widgets.EnumSchemaWidget},
        "enum": {"enum": widgets.EnumSchemaWidget},
        "anyOf": {"anyOf": widgets.AnyOfSchemaWidget, },
    }

    default_widget_variants = {
        "boolean": "checkbox",
        "object": "object",
        "array": "array",
        "number": "spin",
        "integer": "spin",
        "string": "text",
        "enum": "enum",
        "anyOf": "anyOf",
        "directory-path": "dirpath",
    }

    widget_variant_modifiers = {
        "string": lambda schema: schema.get("format", "text")
    }

    def __init__(self, schema: Union[Dict, Schema], validator_cls=None):
        self.widget_map = deepcopy(self.default_widget_map)

        self.schema: Schema = schema if isinstance(schema, Schema) else Schema(schema, validator_cls=validator_cls)
        self.schema.check_schema()

    def create_form(self,
                    ui_schema: Optional[Dict] = None,
                    state: Optional[Dict] = None,
                    parent=None) \
            -> widgets.SchemaWidgetMixin:
        if ui_schema is None:
            ui_schema = {}

        schema_widget = self.create_widget(self.schema.schema, ui_schema, state)
        form = widgets.FormWidget(schema_widget, parent=parent)

        _validator = self.schema.validator()

        def validate(data):
            form.clear_errors()
            errors = [*_validator.iter_errors(data)]

            if errors:
                form.display_errors(errors)

            for err in errors:
                schema_widget.handle_error(err.path, err)

        schema_widget.on_changed.connect(validate)

        return form

    def get_widget_state(self, schema, state=None):
        if state is None:
            return self.schema.compute_defaults(schema)
        return state

    def create_widget(self, schema: dict, ui_schema: dict, state=None, parent=None) -> widgets.SchemaWidgetMixin:
        schema = self.schema.resolve_schema(schema)

        if has_schema_type(schema):
            schema_type = get_schema_type(schema)
        elif is_enum(schema):
            # schema created with pydantic do not have a type key for enum-values, always.
            # Therefore, the schema type "enum" is used in this case.
            schema_type = 'enum'
        elif is_any_of(schema):
            schema_type = 'anyOf'
        elif is_all_of(schema):
            # workaround for interesting allOf-pydantic behaviour
            if len(schema.get('allOf', [])) == 1:
                return self.create_widget(schema['allOf'][0], ui_schema, state=state, parent=parent)
            else:
                raise NotImplementedError('allOf is not supported')

        try:
            default_variant = self.widget_variant_modifiers[schema_type](schema)
        except KeyError:
            default_variant = self.default_widget_variants[schema_type]

        if "enum" in schema:
            default_variant = "enum"

        widget_variant = ui_schema.get('ui:widget', default_variant)
        widget_cls = self.widget_map[schema_type][widget_variant]

        widget = widget_cls(schema, ui_schema, self, parent=parent)

        default_state = self.get_widget_state(schema, state)
        if default_state is not None:
            widget.state = default_state
        return widget



from functools import wraps
from typing import Tuple

from qtpy.QtCore import Signal
from qtpy.QtGui import QColor


class StateProperty(property):

    def setter(self, fset):
        @wraps(fset)
        def _setter(*args):
            *head, value = args
            if value is not None:
                fset(*head, value)

        return super().setter(_setter)


state_property = StateProperty


class SchemaWidgetMixin:
    on_changed = Signal(object)

    VALID_COLOUR = '#ffffff'
    INVALID_COLOUR = '#f6989d'

    def __init__(self, schema: dict, ui_schema: dict, widget_builder: 'IBuilder', **kwargs):
        super().__init__(**kwargs)

        self.schema = schema
        self.ui_schema = ui_schema
        self.widget_builder = widget_builder

        self.on_changed.connect(lambda _: self.clear_error())

        self._show_title: bool = kwargs.get('show_title', True)
        self.configure()

    def configure(self):
        pass

    def show_title(self) -> bool:
        """ show/hide the title in the form-widget """
        return self._show_title

    @state_property
    def state(self):
        raise NotImplementedError(f"{self.__class__.__name__}.state")

    @state.setter
    def state(self, state):
        raise NotImplementedError(f"{self.__class__.__name__}.state")

    def handle_error(self, path: Tuple[str], err: Exception):
        if path:
            raise ValueError("Cannot handle nested error by default")
        self._set_valid_state(err)

    def clear_error(self):
        self._set_valid_state(None)

    def _set_valid_state(self, error: Exception = None):
        palette = self.palette()
        colour = QColor()
        colour.setNamedColor(self.VALID_COLOUR if error is None else self.INVALID_COLOUR)
        palette.setColor(self.backgroundRole(), colour)

        self.setPalette(palette)
        self.setToolTip("" if error is None else error.message)  # TODO



from collections import OrderedDict
from functools import partial
from typing import Tuple, Dict

from qtpy.QtCore import Qt
from qtpy.QtWidgets import QGroupBox, QWidget, QGridLayout, QLabel, QVBoxLayout, QComboBox, QHBoxLayout, QSizePolicy

from pyqtschema.widgets.base import SchemaWidgetMixin, state_property
from pyqtschema.widgets.utils import iter_layout_widgets


class ObjectSchemaWidget(SchemaWidgetMixin, QGroupBox):

    def __init__(self, schema: dict, ui_schema: dict, widget_builder: 'WidgetBuilder', *args, **kwargs):
        super().__init__(schema, ui_schema, widget_builder, *args, **kwargs)

        self.widgets = self.populate_from_schema(schema, ui_schema, widget_builder)

    @state_property
    def state(self) -> dict:
        return {k: w.state for k, w in self.widgets.items()}

    @state.setter
    def state(self, state: dict):
        for name, value in state.items():
            self.widgets[name].state = value

    def handle_error(self, path: Tuple[str], err: Exception):
        name, *tail = path
        self.widgets[name].handle_error(tail, err)

    def widget_on_changed(self, name: str, value):
        self.state[name] = value
        self.on_changed.emit(self.state)

    def populate_from_schema(self, schema: dict, ui_schema: dict, widget_builder: 'WidgetBuilder'
                             ) -> Dict[str, QWidget]:
        # layout = QFormLayout()
        layout = QGridLayout()

        self.setLayout(layout)
        layout.setAlignment(Qt.AlignTop)
        self.setFlat(False)

        if 'title' in schema:
            self.setTitle(schema['title'])

        if 'description' in schema:
            self.setToolTip(schema['description'])

        # Populate rows
        widgets = {}

        for name, sub_schema in schema['properties'].items():
            sub_ui_schema = ui_schema.get(name, {})

            _hide = sub_ui_schema.get('ui:hidden', False)
            _disable = sub_ui_schema.get('ui:disabled', False)
            _show_title = sub_ui_schema.get('ui:show_title', True)

            widget = widget_builder.create_widget(sub_schema, sub_ui_schema, parent=self)  # TODO onchanged
            widget.on_changed.connect(partial(self.widget_on_changed, name))
            widget.setHidden(_hide)
            widget.setDisabled(_disable)

            _row_index = layout.rowCount()
            if widget.show_title() and _show_title:
                label = sub_schema.get("title", name)
                _lbl = QLabel(label, parent=self)
                _lbl.setHidden(_hide)
                _lbl.setDisabled(_disable)

                layout.addWidget(_lbl, _row_index, 0)
                layout.addWidget(widget, _row_index, 1)
                widget: QWidget
                widget.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Minimum)

            else:
                layout.addWidget(widget, _row_index, 0, 1, 2)
            widgets[name] = widget

        return widgets



from .any_of import AnyOfSchemaWidget
from .array import ArraySchemaWidget
from .boolean import CheckboxSchemaWidget
from .color import ColorSchemaWidget
from .enu import EnumSchemaWidget
from .groups import ObjectSchemaWidget
from .numeric import SpinSchemaWidget, SpinDoubleSchemaWidget, IntegerRangeSchemaWidget
from .paths import FilepathSchemaWidget, DirectoryPathSchemaWidget
from .root import FormWidget
from .text import TextSchemaWidget, TextAreaSchemaWidget, PasswordWidget

from .base import SchemaWidgetMixin



from qtpy.QtWidgets import QFileDialog, QHBoxLayout, QLineEdit, QPushButton, QWidget

from pyqtschema.widgets.base import SchemaWidgetMixin, state_property


class FilepathSchemaWidget(SchemaWidgetMixin, QWidget):

    def __init__(self, schema: dict, ui_schema: dict, widget_builder: 'WidgetBuilder', **kwargs):
        super().__init__(schema, ui_schema, widget_builder, **kwargs)

        layout = QHBoxLayout()
        self.setLayout(layout)

        self.path_widget = QLineEdit()
        self.button_widget = QPushButton("Browse")
        layout.addWidget(self.path_widget)
        layout.addWidget(self.button_widget)

        self.button_widget.clicked.connect(self._on_clicked)
        self.path_widget.textChanged.connect(self.on_changed.emit)

    def _on_clicked(self, flag):
        path, filter = QFileDialog.getOpenFileName()
        self.path_widget.setText(path)

    @state_property
    def state(self) -> str:
        return self.path_widget.text()

    @state.setter
    def state(self, state: str):
        self.path_widget.setText(state)


class DirectoryPathSchemaWidget(FilepathSchemaWidget):
    def _on_clicked(self, flag):
        path = QFileDialog.getExistingDirectory()
        self.path_widget.setText(path)



from pyqtschema.schema import Schema


def test_init_schema():
    _def = {
        "title": "Simple",
        "type": "object",
        "properties": {
            "string": {"title": "String", "type": "string"},
            "boolean": {"title": "Boolean", "default": True, "type": "boolean"},
            "integer": {"title": "Integer", "default": 2, "type": "integer"},
            "number": {"title": "Number", "default": 2.2, "type": "number"}
        },
        "required": ["string"]
    }
    schema = Schema(_def)


def test_schema_is_valid():
    _def = {'type': 'object', 'properties': {'checker': {'type': 'string'}},
            'schema': {'additionalProperties': False}, 'additionalProperties': False}
    schema = Schema(_def)

    assert schema.is_valid_data({'checker': 'a string'})
    assert schema.is_valid_data({'checker': ''})

    assert schema.is_valid_data({'checker': 1}) is False
    assert schema.is_valid_data({'checker': 1.2}) is False

    assert schema.is_valid_data({'invalid_name': 1}) is False
    assert schema.is_valid_data({'invalid_name': "1"}) is False



import sys
from enum import Enum
from json import JSONEncoder, dump
from typing import Dict, Any

from qtpy.QtWidgets import QApplication, QScrollArea, QWidget

from pyqtschema.builder import WidgetBuilder


def build_example_widget(schema: Dict, ui_schema=None) -> QWidget:
    builder = WidgetBuilder(schema)
    form = builder.create_form(ui_schema=ui_schema)
    return form


def build_example(schema: Dict, ui_schema: Dict = {}, scrollbar: bool = True):
    app = QApplication(sys.argv)
    widget = form = build_example_widget(schema, ui_schema=ui_schema)
    if scrollbar:
        widget = QScrollArea()
        widget.setWidget(form)
        widget.setWidgetResizable(True)

    form.widget.on_changed.connect(lambda xxx: print(xxx))

    widget.show()
    app.exec_()


class Encoder(JSONEncoder):
    def default(self, o: Any) -> Any:
        if isinstance(o, Enum):
            return o.value
        return JSONEncoder.default(self, o)


def json_dump(obj, fp, *, skipkeys=False, ensure_ascii=True, check_circular=True, allow_nan=True, indent=None,
              separators=None, sort_keys=False, **kwargs):
    """ adjusted json-dump to support Enum-values"""
    kwargs['cls'] = Encoder
    return dump(obj, fp, skipkeys=skipkeys, ensure_ascii=ensure_ascii, check_circular=check_circular,
                allow_nan=allow_nan, indent=indent, separators=separators, sort_keys=sort_keys, **kwargs)
